"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.analyzeVariableDeclaration = exports.analyze = void 0;
const typescript_estree_1 = require("@typescript-eslint/typescript-estree");
const chalk_1 = __importDefault(require("chalk"));
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const config_1 = require("./config");
const constants_1 = require("./constants");
const defaultLineColumn = { line: 1, column: 1 };
const loadedFileMap = new Map();
const analyze = (root, alias, parentPath, parentNode, dir) => {
    try {
        const { fileName } = parentNode;
        let loadedFile;
        if (loadedFileMap.has(fileName)) {
            loadedFile = loadedFileMap.get(fileName);
        }
        else {
            const code = fs_1.default.readFileSync(fileName).toString();
            const parseOptions = {
                loc: true,
                jsx: true,
            };
            const ast = (0, typescript_estree_1.parse)(code, parseOptions);
            loadedFileMap.set(fileName, { code, ast });
        }
        if (!loadedFile) {
            return;
        }
        const { code, ast } = loadedFile;
        const defaultDeclarationName = findDefaultDeclarationName(ast.body);
        if (defaultDeclarationName &&
            parentNode.astType === typescript_estree_1.AST_NODE_TYPES.ImportDefaultSpecifier) {
            parentNode.exportName = defaultDeclarationName;
        }
        for (const state of ast.body) {
            if (!state.type.startsWith('TS')) {
                analyzeAst(root, alias, parentNode, dir, state);
            }
        }
        for (const childNode of parentNode.children) {
            // [Note]: Block circular references.
            if (parentPath !== childNode.fileName &&
                childNode.exists &&
                parentNode.fileName !== childNode.fileName) {
                (0, exports.analyze)(false, alias, fileName, childNode, path_1.default.dirname(childNode.fileName));
            }
        }
    }
    catch (e) {
        console.log(chalk_1.default.red(e));
        parentNode.exists = false;
    }
};
exports.analyze = analyze;
const analyzeAst = (root, alias, parentNode, dir, body) => {
    if (body.type === typescript_estree_1.AST_NODE_TYPES.ImportDeclaration) {
        analyzeImport(alias, parentNode, dir, body);
    }
    else if (body.type === typescript_estree_1.AST_NODE_TYPES.FunctionDeclaration) {
        if (parentNode.astType !== typescript_estree_1.AST_NODE_TYPES.ImportDefaultSpecifier ||
            (parentNode.astType === typescript_estree_1.AST_NODE_TYPES.ImportDefaultSpecifier &&
                body.id &&
                body.id.name === parentNode.id))
            analyzeBlockStatement(parentNode, body.body);
    }
    else if (body.type === typescript_estree_1.AST_NODE_TYPES.ExpressionStatement) {
        if (body.expression.type === typescript_estree_1.AST_NODE_TYPES.CallExpression ||
            body.expression.type === typescript_estree_1.AST_NODE_TYPES.NewExpression) {
            for (const argument of body.expression.arguments) {
                analyzeExpression(parentNode, argument);
            }
        }
    }
    else if (body.type === typescript_estree_1.AST_NODE_TYPES.VariableDeclaration) {
        for (const declaration of body.declarations) {
            if (declaration.id.type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
                const { name } = declaration.id;
                if (declaration.init &&
                    (declaration.init.type === typescript_estree_1.AST_NODE_TYPES.ArrowFunctionExpression ||
                        declaration.init.type === typescript_estree_1.AST_NODE_TYPES.FunctionExpression)) {
                    if (declaration.init.body.type === typescript_estree_1.AST_NODE_TYPES.JSXElement) {
                        parentNode.children.push({
                            id: name,
                            title: name,
                            astType: typescript_estree_1.AST_NODE_TYPES.VariableDeclarator,
                            fileName: parentNode.fileName,
                            // default values
                            exportName: name,
                            coordinates: [0, 0],
                            children: [],
                            exists: false,
                            loc: declaration.id.loc,
                        });
                    }
                    else if (declaration.id.name === parentNode.exportName) {
                        parentNode.loc = declaration.id.loc;
                        analyzeExpression(parentNode, declaration.init);
                    }
                    else if (root) {
                        // [Note]: Root file parsing also parses non-exported declarations.
                        analyzeExpression(parentNode, declaration.init);
                    }
                }
            }
        }
    }
    else if (body.type === typescript_estree_1.AST_NODE_TYPES.ExportNamedDeclaration) {
        if (body.declaration &&
            body.declaration.type === typescript_estree_1.AST_NODE_TYPES.VariableDeclaration) {
            (0, exports.analyzeVariableDeclaration)(parentNode, body.declaration);
        }
    }
};
const analyzeVariableDeclaration = (parentNode, declaration) => {
    if (declaration.type !== typescript_estree_1.AST_NODE_TYPES.VariableDeclaration) {
        return;
    }
    for (const _declaration of declaration.declarations) {
        if (_declaration.id.type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
            if (_declaration.id.name === parentNode.exportName ||
                parentNode.exportName === '') {
                parentNode.loc = _declaration.id.loc;
                if (_declaration.init) {
                    analyzeExpression(parentNode, _declaration.init);
                }
            }
        }
    }
};
exports.analyzeVariableDeclaration = analyzeVariableDeclaration;
const analyzeImport = (alias, parentNode, dir, importDec) => {
    if (importDec.importKind === 'type')
        return;
    const filePath = (0, config_1.resolveAlias)(alias, dir, importDec);
    const fileName = path_1.default.basename(filePath);
    const { filePath: finalFilePath, existsFile } = findAnalyzeFilePath(filePath);
    let id = null;
    if (!existsFile) {
        return;
    }
    for (const specifier of importDec.specifiers) {
        const { type, local } = specifier;
        id = null;
        if (type === typescript_estree_1.AST_NODE_TYPES.ImportDefaultSpecifier &&
            local.type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
            // [Note]: For default import, set the file name.
            id = fileName;
        }
        else if ((type === typescript_estree_1.AST_NODE_TYPES.ImportSpecifier &&
            specifier.importKind === 'value') ||
            (type === typescript_estree_1.AST_NODE_TYPES.ImportNamespaceSpecifier &&
                local.type === typescript_estree_1.AST_NODE_TYPES.Identifier)) {
            id = local.name;
        }
        if (id) {
            parentNode.children.push({
                id,
                title: id,
                astType: type,
                fileName: finalFilePath,
                exportName: local.name,
                // default values
                coordinates: [0, 0],
                children: [],
                exists: false,
                loc: {
                    start: { ...defaultLineColumn },
                    end: { ...defaultLineColumn },
                },
            });
        }
    }
};
const analyzeExpression = (parentNode, elm) => {
    if (!elm) {
        return;
    }
    if (elm.type === typescript_estree_1.AST_NODE_TYPES.JSXElement) {
        analyzeChildrenJSXElement(parentNode, elm);
        updateExists(parentNode, elm.openingElement, parentNode.children);
        analyzeChildrenJSXElement(parentNode, elm);
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.JSXFragment) {
        analyzeChildrenJSXElement(parentNode, elm);
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.SpreadElement) {
        analyzeExpression(parentNode, elm.argument);
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.ArrayExpression) {
        for (const element of elm.elements) {
            analyzeExpression(parentNode, element);
        }
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.AssignmentExpression ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.LogicalExpression) {
        analyzeExpression(parentNode, elm.right);
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.AwaitExpression) {
        analyzeExpression(parentNode, elm.argument);
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.CallExpression) {
        for (const argument of elm.arguments) {
            analyzeExpression(parentNode, argument);
        }
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.ChainExpression) {
        analyzeExpression(parentNode, elm.expression);
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.ClassExpression) {
        // todo
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.ImportExpression) {
        if (elm.attributes) {
            analyzeExpression(parentNode, elm.attributes);
        }
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.MemberExpression) {
        analyzeExpression(parentNode, elm.object);
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.NewExpression) {
        for (const argument of elm.arguments) {
            analyzeExpression(parentNode, argument);
        }
        analyzeExpression(parentNode, elm.callee);
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.ObjectExpression) {
        for (const property of elm.properties) {
            if (property.type === typescript_estree_1.AST_NODE_TYPES.MethodDefinition) {
                if (property.decorators) {
                    for (const decorator of property.decorators) {
                        analyzeExpression(parentNode, decorator.expression);
                    }
                }
                if (property.value.type === typescript_estree_1.AST_NODE_TYPES.FunctionExpression) {
                    analyzeExpression(parentNode, property.value);
                }
            }
        }
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.SequenceExpression) {
        for (const expression of elm.expressions) {
            analyzeExpression(parentNode, expression);
        }
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.TaggedTemplateExpression) {
        for (const expression of elm.quasi.expressions) {
            analyzeExpression(parentNode, expression);
        }
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.TemplateLiteral) {
        for (const expression of elm.expressions) {
            analyzeExpression(parentNode, expression);
        }
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.YieldExpression) {
        if (elm.argument) {
            analyzeExpression(parentNode, elm.argument);
        }
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.ConditionalExpression) {
        analyzeExpression(parentNode, elm.alternate);
        analyzeExpression(parentNode, elm.consequent);
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.ArrowFunctionExpression ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.FunctionExpression) {
        if (elm.body.type === typescript_estree_1.AST_NODE_TYPES.BlockStatement) {
            analyzeBlockStatement(parentNode, elm.body);
        }
        else {
            analyzeExpression(parentNode, elm.body);
        }
    }
    else if (elm.type === typescript_estree_1.AST_NODE_TYPES.ArrayPattern ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.ObjectPattern ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.BinaryExpression ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.Identifier ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.JSXEmptyExpression ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.MetaProperty ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.Super ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.ThisExpression ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.TSAsExpression ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.TSNonNullExpression ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.TSTypeAssertion ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.UnaryExpression ||
        elm.type === typescript_estree_1.AST_NODE_TYPES.UpdateExpression) {
        // [Note]: Unprocessed.
    }
    return;
};
const analyzeBlockStatement = (parentNode, statement) => {
    if (!statement) {
        return;
    }
    for (const b of statement.body) {
        analyzeStatement(parentNode, b);
    }
};
const analyzeStatement = (parentNode, statement) => {
    if (!statement) {
        return;
    }
    if (statement.type === typescript_estree_1.AST_NODE_TYPES.BlockStatement) {
        analyzeBlockStatement(parentNode, statement);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.ClassDeclaration) {
        // todo
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.DoWhileStatement) {
        analyzeStatement(parentNode, statement.body);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.ExpressionStatement) {
        analyzeExpression(parentNode, statement.expression);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.ForInStatement) {
        analyzeStatement(parentNode, statement.body);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.ForOfStatement) {
        analyzeStatement(parentNode, statement.body);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.ForStatement) {
        analyzeStatement(parentNode, statement.body);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.FunctionDeclaration) {
        analyzeBlockStatement(parentNode, statement.body);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.IfStatement) {
        if (statement.alternate) {
            analyzeStatement(parentNode, statement.alternate);
        }
        analyzeStatement(parentNode, statement.consequent);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.ReturnStatement) {
        if (statement.argument) {
            analyzeExpression(parentNode, statement.argument);
        }
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.SwitchStatement) {
        for (const cs of statement.cases) {
            for (const con of cs.consequent) {
                analyzeStatement(parentNode, con);
            }
        }
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.TryStatement) {
        analyzeBlockStatement(parentNode, statement.block);
        if (statement.finalizer) {
            analyzeBlockStatement(parentNode, statement.finalizer);
        }
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.VariableDeclaration) {
        for (const declaration of statement.declarations) {
            if (declaration.init) {
                analyzeExpression(parentNode, declaration.init);
            }
        }
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.WhileStatement) {
        analyzeStatement(parentNode, statement.body);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.WithStatement) {
        analyzeStatement(parentNode, statement.body);
    }
    else if (statement.type === typescript_estree_1.AST_NODE_TYPES.DebuggerStatement ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.ContinueStatement ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.ExportAllDeclaration ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.ExportDefaultDeclaration ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.ExportNamedDeclaration ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.LabeledStatement ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.BreakStatement ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.ThrowStatement ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.TSDeclareFunction ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.TSEnumDeclaration ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.TSExportAssignment ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.TSImportEqualsDeclaration ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.TSInterfaceDeclaration ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.TSModuleDeclaration ||
        statement.type === typescript_estree_1.AST_NODE_TYPES.TSNamespaceExportDeclaration) {
        // [Note]: Unprocessed.
    }
};
const analyzeChildrenJSXElement = (parentNode, element) => {
    for (const child of element.children) {
        if (child.type === typescript_estree_1.AST_NODE_TYPES.JSXElement ||
            child.type === typescript_estree_1.AST_NODE_TYPES.JSXFragment) {
            analyzeExpression(parentNode, child);
        }
        else if (child.type === typescript_estree_1.AST_NODE_TYPES.JSXExpressionContainer ||
            child.type === typescript_estree_1.AST_NODE_TYPES.JSXSpreadChild) {
            analyzeExpression(parentNode, child.expression);
        }
        else if (child.type === typescript_estree_1.AST_NODE_TYPES.JSXEmptyExpression ||
            child.type == typescript_estree_1.AST_NODE_TYPES.JSXText) {
            // [Note]: Unprocessed.
        }
    }
};
const updateExists = (parentNode, openingElm, children) => {
    if (openingElm.name.type === typescript_estree_1.AST_NODE_TYPES.JSXIdentifier) {
        const name = openingElm.name.name;
        const child = children.find((child) => child.exportName === name);
        if (!child) {
            return;
        }
        child.exists = true;
        parentNode.exists = true;
    }
    else if (openingElm.name.type === typescript_estree_1.AST_NODE_TYPES.JSXNamespacedName) {
        if (openingElm.name.name.type === typescript_estree_1.AST_NODE_TYPES.JSXIdentifier) {
            const name = openingElm.name.name.name;
            const child = children.find((child) => child.exportName === name);
            if (!child) {
                return;
            }
            child.exists = true;
            parentNode.exists = true;
        }
    }
    else if (openingElm.name.type === typescript_estree_1.AST_NODE_TYPES.JSXMemberExpression) {
        if (openingElm.name.object.type === typescript_estree_1.AST_NODE_TYPES.JSXIdentifier) {
            const name = openingElm.name.object.name;
            const childIndex = children.findIndex((child) => child.exportName === name);
            if (childIndex === -1) {
                return;
            }
            const child = children[childIndex];
            if (openingElm.name.property.type === typescript_estree_1.AST_NODE_TYPES.JSXIdentifier) {
                child.exists = false;
                const { name } = openingElm.name.property;
                const childFromProperty = {
                    id: name,
                    astType: typescript_estree_1.AST_NODE_TYPES.ImportSpecifier,
                    title: name,
                    fileName: child.fileName,
                    exportName: name,
                    coordinates: [0, 0],
                    children: [],
                    exists: true,
                    loc: {
                        start: { ...defaultLineColumn },
                        end: { ...defaultLineColumn },
                    },
                };
                children.splice(childIndex, 0, childFromProperty);
                parentNode.exists = true;
            }
        }
    }
    return;
};
const findAnalyzeFilePath = (basePath) => {
    for (const ext of constants_1.EXTENTIONS) {
        if (fs_1.default.existsSync(`${basePath}${ext}`)) {
            return { filePath: `${basePath}${ext}`, existsFile: true };
        }
        else if (fs_1.default.existsSync(`${basePath}/index${ext}`)) {
            return { filePath: `${basePath}/index${ext}`, existsFile: true };
        }
    }
    return { filePath: '', existsFile: false };
};
// [Note]: The calling component cannot get the exact name of the loaded function by default.
const findDefaultDeclarationName = (body) => {
    const defaultDeclaration = body.find(({ type }) => type === typescript_estree_1.AST_NODE_TYPES.ExportDefaultDeclaration);
    if (!defaultDeclaration) {
        return undefined;
    }
    if (defaultDeclaration.type === typescript_estree_1.AST_NODE_TYPES.ExportDefaultDeclaration) {
        const { type } = defaultDeclaration.declaration;
        if (type === typescript_estree_1.AST_NODE_TYPES.Identifier) {
            const { name } = defaultDeclaration.declaration;
            return name;
        }
    }
};
